@@ -7,11 +7,11 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/gruntwork-io/terragrunt/config"
-	"github.com/gruntwork-io/terragrunt/options"
+	"github.com/gruntwork-io/terragrunt/internal/tf"
+	"github.com/gruntwork-io/terragrunt/internal/util"
+	"github.com/gruntwork-io/terragrunt/pkg/config"
 	"github.com/gruntwork-io/terragrunt/pkg/log"
-	"github.com/gruntwork-io/terragrunt/tf"
-	"github.com/gruntwork-io/terragrunt/util"
+	"github.com/gruntwork-io/terragrunt/pkg/options"
 )
 
 const (
@@ -24,6 +24,7 @@ type Unit struct {
 	discoveryContext *DiscoveryContext
 	Execution        *UnitExecution
 	path             string
+	configFile       string
 	reading          []string
 	dependencies     Components
 	dependents       Components
@@ -45,12 +46,22 @@ type UnitExecution struct {
 func NewUnit(path string) *Unit {
 	return &Unit{
 		path:             path,
+		configFile:       config.DefaultTerragruntConfigPath,
 		discoveryContext: &DiscoveryContext{},
 		dependencies:     make(Components, 0),
 		dependents:       make(Components, 0),
 	}
 }
 
+// NewUnitExecution creates a new UnitExecution with the given options.
+func NewUnitExecution(l log.Logger, opts *options.TerragruntOptions, excluded bool) *UnitExecution {
+	return &UnitExecution{
+		TerragruntOptions: opts,
+		Logger:            l,
+		FlagExcluded:      excluded,
+	}
+}
+
 // WithReading appends a file to the list of files being read by this component.
 // Useful for constructing components with all files read at once.
 func (u *Unit) WithReading(files ...string) *Unit {
@@ -83,6 +94,16 @@ func (u *Unit) StoreConfig(cfg *config.TerragruntConfig) {
 	u.cfg = cfg
 }
 
+// ConfigFile returns the discovered config filename for this unit.
+func (u *Unit) ConfigFile() string {
+	return u.configFile
+}
+
+// SetConfigFile sets the discovered config filename for this unit.
+func (u *Unit) SetConfigFile(filename string) {
+	u.configFile = filename
+}
+
 // Kind returns the kind of component (always Unit for Unit).
 func (u *Unit) Kind() Kind {
 	return UnitKind
@@ -147,6 +168,15 @@ func (u *Unit) SetDiscoveryContext(ctx *DiscoveryContext) {
 	u.discoveryContext = ctx
 }
 
+// Origin returns the origin of the discovery context for this component.
+func (u *Unit) Origin() Origin {
+	if u.discoveryContext == nil {
+		return OriginUnknown
+	}
+
+	return u.discoveryContext.Origin()
+}
+
 // lock locks the Unit.
 func (u *Unit) lock() {
 	u.mu.Lock()
@@ -336,7 +366,10 @@ func (u *Unit) planFilePath(opts *options.TerragruntOptions, outputFolder, fileN
 		return ""
 	}
 
-	relPath, err := filepath.Rel(opts.RootWorkingDir, u.path)
+	// Use discoveryContext.WorkingDir as base (always populated).
+	// This is critical for git-based filters where units are discovered in temporary worktrees.
+	// Using opts.RootWorkingDir would cause relative paths to escape the outputFolder.
+	relPath, err := filepath.Rel(u.discoveryContext.WorkingDir, u.path)
 	if err != nil {
 		relPath = u.path
 	}
