@@ -1,10 +1,10 @@
 package filter
 
 import (
+	"fmt"
 	"path/filepath"
 	"slices"
 
-	"github.com/gobwas/glob"
 	"github.com/gruntwork-io/terragrunt/internal/component"
 	"github.com/gruntwork-io/terragrunt/pkg/log"
 )
@@ -26,6 +26,14 @@ const (
 	MaxTraversalDepth = 1000000
 )
 
+// graphTraversalParams consolidates parameters for filter graph traversal.
+type graphTraversalParams struct {
+	resultSet   map[string]component.Component
+	visited     map[string]int
+	direction   GraphDirection
+	warnOnLimit bool
+}
+
 // EvaluationContext provides additional context for filter evaluation, such as Git worktree directories.
 type EvaluationContext struct {
 	// GitWorktrees maps Git references to temporary worktree directory paths.
@@ -62,21 +70,16 @@ func Evaluate(l log.Logger, expr Expression, components component.Components) (c
 
 // evaluatePathFilter evaluates a path filter using glob matching.
 func evaluatePathFilter(filter *PathExpression, components component.Components) (component.Components, error) {
-	g, err := filter.CompileGlob()
-	if err != nil {
-		return nil, NewEvaluationErrorWithCause("failed to compile glob pattern: "+filter.Value, err)
-	}
-
 	result := make(component.Components, 0, len(components))
 
-	for _, component := range components {
-		matches, err := doesComponentValueMatchGlob(component, component.Path(), filter.Value, g)
+	for _, c := range components {
+		matches, err := matchPath(c, filter)
 		if err != nil {
-			return nil, err
+			return nil, NewEvaluationErrorWithCause("failed to match path pattern: "+filter.Value, err)
 		}
 
 		if matches {
-			result = append(result, component)
+			result = append(result, c)
 		}
 	}
 
@@ -156,7 +159,7 @@ func evaluateAttributeFilter(filter *AttributeExpression, components []component
 			for _, reading := range c.Reading() {
 				rel, err := filepath.Rel(c.DiscoveryContext().WorkingDir, reading)
 				if err != nil {
-					return nil, NewEvaluationErrorWithCause("failed to get relative path: "+reading, err)
+					return nil, NewEvaluationErrorWithCause(fmt.Sprintf("failed to get relative path for component %s reading: %s", c.Path(), reading), err)
 				}
 
 				relReading = append(relReading, filepath.ToSlash(rel))
@@ -195,20 +198,32 @@ func evaluatePrefixExpression(l log.Logger, expr *PrefixExpression, components c
 		return nil, err
 	}
 
-	excludeSet := make(map[string]struct{}, len(toExclude))
-	for _, c := range toExclude {
-		excludeSet[c.Path()] = struct{}{}
+	if len(toExclude) == 0 {
+		return components, nil
 	}
 
-	var result component.Components
+	// Build a set of paths to exclude for efficient lookup.
+	// We compare by path rather than object identity because graph traversal
+	// may return component instances from Dependencies()/Dependents() that are
+	// different objects than those in the input list.
+	excludePaths := make(map[string]struct{}, len(toExclude))
+	for _, c := range toExclude {
+		excludePaths[c.Path()] = struct{}{}
+	}
+
+	// We don't use slices.DeleteFunc here because we don't want the members of the original components slice to be
+	// zeroed.
+	results := make(component.Components, 0, len(components)-len(toExclude))
 
 	for _, c := range components {
-		if _, ok := excludeSet[c.Path()]; !ok {
-			result = append(result, c)
+		if _, excluded := excludePaths[c.Path()]; excluded {
+			continue
 		}
+
+		results = append(results, c)
 	}
 
-	return result, nil
+	return results, nil
 }
 
 // evaluateInfixExpression evaluates an infix expression (intersection).
@@ -237,6 +252,13 @@ func evaluateGraphExpression(l log.Logger, expr *GraphExpression, components com
 		return nil, err
 	}
 
+	// NOTE: We previously filtered out components with OriginGraphDiscovery here to avoid
+	// including components that were only discovered via graph relationships. However, this
+	// caused issues with intersection filters like "service... | !^db..." where db is
+	// discovered via the first filter and then needs to be used as a target in the second.
+	// The discovery phase already handles this logic properly, so we don't need to filter
+	// by origin here during filter evaluation.
+
 	if len(targetMatches) == 0 {
 		return component.Components{}, nil
 	}
@@ -249,19 +271,45 @@ func evaluateGraphExpression(l log.Logger, expr *GraphExpression, components com
 		}
 	}
 
-	visited := make(map[string]bool)
-
 	if expr.IncludeDependencies {
+		depth := MaxTraversalDepth
+		warnOnLimit := true
+
+		if expr.DependencyDepth > 0 {
+			depth = expr.DependencyDepth
+			warnOnLimit = false
+		}
+
+		params := &graphTraversalParams{
+			resultSet:   resultSet,
+			visited:     make(map[string]int),
+			direction:   GraphDirectionDependencies,
+			warnOnLimit: warnOnLimit,
+		}
+
 		for _, target := range targetMatches {
-			traverseDependencies(l, target, resultSet, visited, MaxTraversalDepth)
+			traverseGraph(l, target, params, depth)
 		}
 	}
 
-	visited = make(map[string]bool)
-
 	if expr.IncludeDependents {
+		depth := MaxTraversalDepth
+		warnOnLimit := true
+
+		if expr.DependentDepth > 0 {
+			depth = expr.DependentDepth
+			warnOnLimit = false
+		}
+
+		params := &graphTraversalParams{
+			resultSet:   resultSet,
+			visited:     make(map[string]int),
+			direction:   GraphDirectionDependents,
+			warnOnLimit: warnOnLimit,
+		}
+
 		for _, target := range targetMatches {
-			traverseDependents(l, target, resultSet, visited, MaxTraversalDepth)
+			traverseGraph(l, target, params, depth)
 		}
 	}
 
@@ -292,20 +340,26 @@ func evaluateGitFilter(filter *GitExpression, components component.Components) (
 	return results, nil
 }
 
-// traverseDependencies recursively traverses the dependency graph downward (from a component to its dependencies).
-func traverseDependencies(
+// traverseGraph recursively traverses the graph in the specified direction (dependencies or dependents).
+// The visited map tracks the maximum remaining depth at which each node was visited, allowing re-traversal
+// when a node is reached with more remaining depth (e.g., from a closer target).
+// The warnOnLimit flag controls whether to log a warning when depth is exhausted (used for safety limits only).
+func traverseGraph(
 	l log.Logger,
 	c component.Component,
-	resultSet map[string]component.Component,
-	visited map[string]bool,
-	maxDepth int,
+	params *graphTraversalParams,
+	remainingDepth int,
 ) {
-	if maxDepth <= 0 {
-		if l != nil {
+	if remainingDepth <= 0 {
+		if l != nil && params.warnOnLimit {
+			directionName := params.direction.String()
+
 			l.Warnf(
-				"Maximum dependency traversal depth (%d) reached for component %s during filtering. Some dependencies may have been excluded from results.",
+				"Maximum %s traversal depth (%d) reached for component %s during filtering. Some %s may have been excluded from results.",
+				directionName,
 				MaxTraversalDepth,
 				c.Path(),
+				directionName,
 			)
 		}
 
@@ -313,82 +367,45 @@ func traverseDependencies(
 	}
 
 	path := c.Path()
-	if visited[path] {
+
+	if prevDepth, seen := params.visited[path]; seen && prevDepth >= remainingDepth {
 		return
 	}
 
-	visited[path] = true
+	params.visited[path] = remainingDepth
 
-	for _, dep := range c.Dependencies() {
-		depPath := dep.Path()
-		resultSet[depPath] = dep
+	var relatedComponents []component.Component
+	if params.direction == GraphDirectionDependencies {
+		relatedComponents = c.Dependencies()
+	} else {
+		relatedComponents = c.Dependents()
+	}
 
-		traverseDependencies(l, dep, resultSet, visited, maxDepth-1)
+	for _, related := range relatedComponents {
+		relatedPath := related.Path()
+
+		// It's not clear why this isn't necessary. It might be in the future.
+		// Tests pass without it, however, so we'll leave it out for now.
+		//
+		// Needs more investigation.
+		//
+		// relatedCtx := related.DiscoveryContext()
+		// if relatedCtx != nil {
+		// 	origin := relatedCtx.Origin()
+		// 	if origin != component.OriginGraphDiscovery {
+		// 		l.Debugf(
+		// 			"Skipping %s %s in graph expression traversal: component was discovered via %s, not graph discovery",
+		// 			direction.String(),
+		// 			relatedPath,
+		// 			origin,
+		// 		)
+
+		// 		continue
+		// 	}
+		// }
+
+		params.resultSet[relatedPath] = related
+
+		traverseGraph(l, related, params, remainingDepth-1)
 	}
 }
-
-// traverseDependents recursively traverses the dependent graph upward (from a component to its dependents).
-func traverseDependents(
-	l log.Logger,
-	c component.Component,
-	resultSet map[string]component.Component,
-	visited map[string]bool,
-	maxDepth int,
-) {
-	if maxDepth <= 0 {
-		if l != nil {
-			l.Warnf(
-				"Maximum dependent traversal depth (%d) reached for component %s during filtering. Some dependents may have been excluded from results.",
-				MaxTraversalDepth,
-				c.Path(),
-			)
-		}
-
-		return
-	}
-
-	path := c.Path()
-	if visited[path] {
-		return
-	}
-
-	visited[path] = true
-
-	for _, dependent := range c.Dependents() {
-		depPath := dependent.Path()
-		resultSet[depPath] = dependent
-
-		traverseDependents(l, dependent, resultSet, visited, maxDepth-1)
-	}
-}
-
-// doesComponentValueMatchGlob checks if a component matches a glob pattern.
-//
-// It does some logic to assess whether the original value that was compiled into a glob was an absolute path or not,
-// and uses that to determine how to match the component.Path() against the glob.
-//
-// If it's an absolute path, it matches the component.Path() against the glob as is.
-// If it's not an absolute path, it attempts to translate the component.Path() path into a relative path
-// using the discovery context's working directory, and then matches the relative path against the glob.
-//
-// If neither are true, it does a best effort match attempting to match the component.Path() against the value as is.
-func doesComponentValueMatchGlob(c component.Component, val, globVal string, glob glob.Glob) (bool, error) {
-	if filepath.IsAbs(globVal) {
-		return glob.Match(filepath.ToSlash(val)), nil
-	}
-
-	discoveryCtx := c.DiscoveryContext()
-	if discoveryCtx != nil &&
-		discoveryCtx.WorkingDir != "" {
-		relPath, err := filepath.Rel(discoveryCtx.WorkingDir, val)
-		if err != nil {
-			return false, NewEvaluationErrorWithCause("failed to get relative path: "+val, err)
-		}
-
-		relPath = filepath.ToSlash(relPath)
-
-		return glob.Match(relPath), nil
-	}
-
-	return glob.Match(filepath.ToSlash(val)), nil
-}
